import atexit
import datetime
import json
import os
import re
import requests
import shutil
import tempfile
import time
import traceback

from requests.exceptions import ConnectionError
from requests.auth import _basic_auth_str


from .conda_interface import (
    download,
    fetch_index,
    known_subdirs,
    VersionOrder,
    empty_repodata,
)
from .config import config, parse_argv
from .filter import filter_all
from .indices import (
    index_keys,
    channel_keys,
    build_subdir_indices,
    build_channel_indices,
)
from .utils import compute_checksums, MirrorException, count_pkgs
from .logging import logger as _l, set_log_level, log_prefix, is_verbose


# Remove these keys at the destination
keys_to_remove = ("fn", "channel", "binstar", "target-triplet")
# If any of these keys change, we must upload a new package
keys_to_replace = {"size": -1, "md5": "", "sha256": "", "sha1": ""}
# If any of these keys change, and none of the above, we can just patch
keys_to_patch = {
    "name": "",
    "version": "",
    "build": "",
    "build_number": -1,
    "timestamp": -1,
    "depends": [],
    "constrains": [],
    "features": "",
    "track_features": "",
    "license": "",
    "license_family": "",
}


def _retry(action):
    sleep_time = 1.0
    for attempt in range(config.max_attempts):
        try:
            return action()
        except Exception as exc:
            _l.warning("! Connection error: %s", exc)
            if attempt == config.max_attempts - 1:
                _l.error("! %d attempts exhausted", config.max_attempts)
                raise MirrorException("Download failed")
        _l.warning(
            "! Retrying in %g second%s", sleep_time, "" if sleep_time == 1 else "s"
        )
        time.sleep(sleep_time)
        sleep_time *= 2.0


_session = None


def get_session():
    global _session
    if _session is None:
        _session = requests.Session()
        _session.headers.update({"Connection": "Keep-Alive"})
        if config.password:
            if config.username:
                auth_str = _basic_auth_str(config.username, config.password)
            else:
                auth_str = "Bearer " + config.password
            _session.headers["Authorization"] = auth_str
        if getattr(config, "api_headers", None):
            _session.headers.update(config.api_headers)
        if config.api_url.startswith("https"):
            _session.verify = config.verify_ssl_mode("dst")
    return _session


def _decode(b):
    if not b:
        return ""
    elif not isinstance(b, bytes):
        return b
    try:
        return b.decode("utf-8")
    except UnicodeDecodeError:
        return b.decode("iso-8859-1")


def _safeurl(b):
    return re.sub(r"(^|.*/)t/[^/]*(.*)", r"\1t/<TOKEN>\2", b)


def _canonicalize(url, channel_alias):
    if not (url.startswith((".", "/")) or "://" in url):
        url = channel_alias + url
    if url.startswith(channel_alias):
        return url, url[len(channel_alias) :]
    elif url.startswith("file://"):
        return url, url[7:]
    return url, url


CONTENT_FLAGS = (
    (".json", {"Content-Type": "application/json; charset=utf-8"}),
    (
        ".json.gz",
        {"Content-Type": "application/json; charset=utf-8", "Content-Encoding": "gzip"},
    ),
    ("", {"Content-Type": "application/octet-stream"}),
)


def _content_data(fn):
    for sfx, cdata in CONTENT_FLAGS:
        if not sfx or fn.endswith(sfx):
            return cdata.copy()


def _api(method, *urlparts, **kwargs):
    method = method.upper()
    short_url = "/".join(p for p in urlparts if p)
    if short_url.startswith("/"):
        url = "/".join(config.api_url.split("/")[:3]) + short_url
    else:
        url = config.api_url.rstrip("/") + "/" + short_url
    throw = kwargs.pop("throw", True)
    accept_404 = kwargs.pop("accept_404", False)
    headers = kwargs.pop("headers", None) or {}
    if "json" in kwargs:
        headers.setdefault("Content-Type", "application/json")
        kwargs["data"] = json.dumps(kwargs.pop("json"))
    if headers:
        kwargs["headers"] = headers
    sleep_time = 1.0
    session = get_session()
    for attempt in range(config.max_attempts):
        try:
            response = session.request(method, url, **kwargs)
            reason = _decode(getattr(response, "reason", ""))
            if not response.content:
                data = None
            elif "json" in response.headers.get("Content-Type", ""):
                data = response.json()
                if "jsonapi" in data and "data" in data:
                    data = data["data"]
            else:
                data = _decode(response.content)
            reason = [reason]
            if isinstance(data, dict):
                reason.append(data.get("message"))
                if isinstance(data.get("errors"), list):
                    reason.extend(e.get("message") for e in data["errors"])
            reason = [r for r in reason if r]
            message = "%s %s %d" % (method, _safeurl(short_url), response.status_code)
            if reason:
                message += " " + "\n> ".join(reason)
            if response.status_code < 300 or accept_404 and response.status_code == 404:
                for mstr in message.splitlines():
                    _l.debug("%s", message)
                return data
            break
        except ConnectionError as exc:
            _l.warning("! %s", exc)
            if attempt >= config.max_attempts - 1:
                _l.error("! %d retry attempts exhausted", config.max_attempts)
                message = "Persisent connection error"
                break
            _l.warning(
                "! retrying in %g second%s",
                sleep_time,
                "" if sleep_time == 1 else "s",
            )
            time.sleep(sleep_time)
            sleep_time *= 2.0
    exc = MirrorException(message)
    if throw:
        raise exc
    return exc


_tmpdir = None


def get_tmpdir():
    global _tmpdir
    if _tmpdir is None:
        _tmpdir = tempfile.mkdtemp()
        atexit.register(shutil.rmtree, _tmpdir, ignore_errors=False)
        _l.debug("Temporary directory created.")
    return _tmpdir


def verify_checksums(info_src, path):
    info_comp = compute_checksums(path)
    for name in ("Size", "MD5", "SHA1", "SHA256"):
        key = name.lower()
        sval = info_comp.get(key)
        dval = info_src.get(key)
        if dval in (0, None):
            info_src[key] = sval
        elif sval != dval:
            _l.error("! %s mismatch: download %s, metadata %s", name, sval, dval)
            raise MirrorException("Checksum mismatch")


def transmute(info_src, path):
    from conda_package_handling.api import transmute as cph_transmute

    fname = os.path.basename(path)
    if fname.endswith("conda"):
        sfmt = "conda"
        dfmt = "tar.bz2"
    else:
        sfmt = "tar.bz2"
        dfmt = "conda"
    fname2 = fname[: -len(sfmt)] + dfmt
    path2 = os.path.join(get_tmpdir(), fname2)
    failed_files = cph_transmute(path, "." + dfmt, get_tmpdir())
    if failed_files:
        msg = list(failed_files.values())[0]
        _l.error("Conversion failed: %s", msg)
        raise MirrorException("Conversion failed")
    info_src.pop("fn", None)
    orec = info_src["_orig"] = {"_format": sfmt}
    for k in keys_to_replace:
        if k in info_src:
            orec[k] = info_src[k]
            del info_src[k]
    return path2


def _format_keys(fn):
    if fn in index_keys or fn in channel_keys:
        fmt = "index"
        pkey = "indices"
    elif fn.endswith(".conda"):
        fmt = "conda"
        pkey = "packages.conda"
    elif fn.endswith(".tar.bz2"):
        fmt = "tar.bz2"
        pkey = "packages"
    else:
        fmt = "extra"
        pkey = "extras"
    return fmt, pkey


def reconcile_packages(repodata, packages, ignore_missing=False):
    if not repodata or packages is None:
        return False
    counts = {"unregistered": 0, "missing": 0, "corrupt": 0}
    repodata["indices"] = {}
    repodata["extras"] = {}
    all_names = set(packages)
    all_names.update(repodata["packages"])
    all_names.update(repodata["packages.conda"])
    for fn in all_names:
        srec = packages.get(fn)
        _, pkey = _format_keys(fn)
        prec = repodata[pkey].get(fn)
        if srec is None:
            counts["missing"] += 1
            if not ignore_missing:
                del repodata[pkey][fn]
        elif prec is None:
            repodata[pkey][fn] = srec
            if pkey != "indices":
                counts["unregistered"] += 1
        elif any(v != prec[k] for k, v in srec.items() if k in prec):
            counts["corrupt"] += 1
            prec.update(srec)
    if any(counts.values()):
        if ignore_missing:
            counts["missing (ignored)"] = counts["missing"]
            counts["missing"] = 0
        msg = "found " + ", ".join("%d %s" % (v, k) for k, v in counts.items() if v)
        return msg


def _call(method, module, *args, **kwargs):
    func = getattr(module, method, None)
    if func is None:
        return
    with log_prefix("| "):
        try:
            return func(*args, **kwargs)
        except MirrorException as exc:
            message = getattr(exc, "message", None) or str(exc)
            message = message.splitlines()
            _l.error("ERROR in %s: %s", method, message[0])
            for estr in message[1:]:
                _l.error(estr)
            raise


def sync_files(
    module,
    api,
    cname,
    platform,
    sorted_keys,
    src_pkgs,
    dst_pkgs,
    ignore_missing=False,
    can_patch=False,
    printed=False,
):
    dpath = (module, api, cname)
    if platform:
        dpath = dpath + (platform,)
    pkg_counts = {"before": len(dst_pkgs)}
    while sorted_keys:
        # We cannot just iterate through the list because the transmute
        # logic must in some cases add an additional fn to the search.
        fn = src_fn = sorted_keys.pop(0)
        fmt, fmt_key = _format_keys(fn)
        is_package = fmt_key.startswith("package")

        info_src = src_pkgs.get(fn)
        info_dst = dst_pkgs.get(fn)
        skip = info_src.get("_skip") if info_src else None
        # This is so that we can detect transmutation cases
        src_fn = info_src["fn"] if info_src and "fn" in info_src else fn
        repl_dst = info_dst.get("_orig", {}) if info_dst and src_fn != fn else info_dst

        if not info_src and not info_dst:
            continue
        elif skip and info_src and not info_dst:
            if skip == "transmuting":
                action, action_p = "Transmuting", "silent"
            elif skip == "redundant":
                action, action_p = "Redundant", "silent"
            elif skip == "format":
                action, action_p = "Blocking (format)", "blocked"
            else:
                action, action_p = "Skipping (%s)" % skip, "filtered"
        elif not info_src or skip:
            # Package is on the destination but not the source
            sfx = " (%s)" % skip if skip else ""
            if config.clean:
                action, action_p = "Removing%s" % sfx, "removed"
            else:
                action, action_p = "Keeping%s" % sfx, "kept"
        elif not info_dst:
            # Package is on the source but not the destination
            if src_fn == fn:
                action, action_p = "Adding", "added"
            else:
                action, action_p = "Converting", "added"
        elif any(
            info_src[k] != repl_dst[k]
            for k in keys_to_replace
            if k in info_src and k in repl_dst
        ):
            # Package is present in both, but the file itself is different
            action, action_p = "Replacing", "replaced"
        elif fmt not in ("index", "extra") and any(
            info_src.get(k, v) != info_dst.get(k, v) for k, v in keys_to_patch.items()
        ):
            if can_patch:
                action, action_p = "Patching", "patched"
            else:
                action, action_p = "Replacing", "replaced"
        else:
            action, action_p = "Up to date", "untouched"
            if src_fn != fn:
                action += " (converted)"

        skipv = action_p in ("filtered", "untouched", "transmuting", "kept")
        sfx = ""
        if action in ("Patching", "Replacing"):
            pkeys = []
            if info_src.get("size") != info_dst.get("size"):
                pkeys.append("content")
            if not pkeys:
                pkeys.extend(
                    k
                    for k in keys_to_replace
                    if k in info_src and k in info_dst and info_src[k] != info_dst[k]
                )
            if not pkeys:
                pkeys.extend(
                    k
                    for k in keys_to_patch
                    if k in info_src and info_src[k] != info_dst.get(k)
                )
                if len(pkeys) * 2 >= len(keys_to_patch):
                    pkeys = ["metadata"]
            sfx += " (%s)" % (",".join(pkeys))
        if not skipv and config.dry_run:
            sfx += " (dry run)"
        if action_p != "silent" and (not skipv or is_verbose()):
            level = "debug" if skipv else "info"
            getattr(_l, level)("%s: %s%s", fn, action, sfx)
            printed = True

        paths_to_remove = []
        fpath = dpath + (fn,)
        try:
            if config.dry_run:
                pass

            elif action_p == "removed":
                _call("del_object", *fpath, metadata=info_dst)
                if is_package:
                    del dst_pkgs[fn]

            elif action_p == "patched":
                info_dst.update(info_src)
                info_dst = {
                    k: v for k, v in info_dst.items() if k not in keys_to_remove
                }
                _call("patch_package", *fpath, metadata=info_dst)
                if is_package:
                    dst_pkgs[fn] = info_dst

            elif action_p not in ("added", "converted", "replaced"):
                pass

            elif not fmt_key.startswith("packages"):
                path = info_src.pop("_data")
                info_dst = {
                    k: v for k, v in info_src.items() if k not in keys_to_remove
                }
                _call("put_object", *fpath, path_or_data=path, metadata=info_dst)

            else:
                url = info_src["channel"].rstrip("/") + "/" + src_fn
                url, short_url = _canonicalize(url, config.channel_alias)
                if url.startswith("file://"):
                    _l.debug("| Local: %s", _safeurl(short_url))
                    path = url[7:]
                else:
                    _l.debug("| Fetch: %s", _safeurl(short_url))
                    path = os.path.join(get_tmpdir(), src_fn)
                    paths_to_remove.append(path)
                    _retry(lambda: download(url, path))
                info_src = {
                    k: v for k, v in info_src.items() if k not in keys_to_remove
                }
                verify_checksums(info_src, path)
                if src_fn != fn:
                    _l.debug("%s -> %s", src_fn, fn)
                    path = transmute(info_src, path)
                    paths_to_remove.append(path)
                    verify_checksums(info_src, path)
                if action_p == "replaced":
                    _call("del_object", *fpath, metadata=info_dst)
                _call("put_object", *fpath, path_or_data=path, metadata=info_src)
                if is_package:
                    dst_pkgs[fn] = info_src

        except MirrorException:
            config.error_count += 1
            action_p = "errored"
            fn = (fn, action)

        finally:
            while paths_to_remove:
                path = paths_to_remove.pop()
                if os.path.exists(path):
                    os.remove(path)

        if action_p != "silent" and fmt_key.startswith("packages"):
            pkg_counts[action_p] = pkg_counts.get(action_p, 0) + 1
        if config.error_count >= config.max_failures:
            _l.error("The maximum number of errors have been reached.")
            break

    pkg_counts["after"] = len(dst_pkgs)
    if not printed:
        _l.info("| up to date")
    return pkg_counts


def retrieve_platform_source(module, api, cname, platform):
    src_repodata = fetch_index(
        config.channels, platform, config.channel_alias, config.verify_ssl_mode("src")
    )
    src_pkgs = {**src_repodata["packages"], **src_repodata["packages.conda"]}
    n_blobs, n_pkgs = count_pkgs(src_pkgs)
    if n_blobs or n_pkgs or is_verbose():
        _l.info("%s: %d blobs, %d packages", platform, n_blobs, n_pkgs)
    return src_pkgs


def retrieve_sources(module, api, cname):
    all_sources = {}
    _l.info("Retrieving source metadata")
    with log_prefix("  "):
        for platform in config.platforms or known_subdirs():
            src_pkgs = retrieve_platform_source(module, api, cname, platform)
            if src_pkgs:
                all_sources[platform] = src_pkgs
    if not all_sources:
        _l.info("| no source packages found")
    return all_sources


def retrieve_platform_destination(module, api, cname, platform):
    dpath = (module, api, cname, platform)
    skip_platform = config.platforms and platform not in config.platforms
    need_indices = hasattr(module, "get_object")
    if need_indices:
        dst_repodata = _call("get_object", *dpath, "repodata.json")
        if dst_repodata and isinstance(dst_repodata, (str, bytes)):
            dst_repodata = json.loads(dst_repodata)
    else:
        dst_repodata = fetch_index(
            config.dest_channel,
            platform,
            config.dest_channel_alias,
            config.verify_ssl_mode("dst"),
        )
    if not dst_repodata:
        dst_repodata = empty_repodata(platform)
    dst_objects = _call("list_objects", *dpath, repodata=dst_repodata)
    if dst_objects is None:
        if skip_platform:
            return {}, {}
        found_orphans = None
    else:
        found_orphans = reconcile_packages(
            dst_repodata, dst_objects, config.ignore_missing
        )
    dst_pkgs = {**dst_repodata["packages"], **dst_repodata["packages.conda"]}
    n_blobs, n_pkgs = count_pkgs(dst_pkgs)
    if n_blobs or n_pkgs or found_orphans or is_verbose():
        _l.info("%s: %d blobs, %d packages", platform, n_blobs, n_pkgs)
        if found_orphans:
            _l.info("| %s", found_orphans)
    return dst_pkgs, dst_repodata


def retrieve_destinations(module, api, cname, all_sources=None):
    all_destinations = {}
    _l.info("Retrieving destination metadata")
    with log_prefix("  "):
        found = False
        all_platforms = known_subdirs()
        if config.clean:
            dst_platforms = all_platforms
        else:
            dst_platforms = config.platforms or all_platforms
        for platform in dst_platforms:
            if not config.clean and all_sources is not None:
                if platform not in all_sources:
                    _l.info("| %s: skipping (no packages)", platform)
                continue
            dst_pkgs, dst_repodata = retrieve_platform_destination(
                module, api, cname, platform
            )
            all_destinations[platform] = (dst_pkgs, dst_repodata)
            if dst_pkgs or dst_repodata:
                found = True
        if not found:
            _l.info("| no packages found")
    return all_destinations


def sync_platform(module, api, cname, platform, src_pkgs, dst_pkgs, dst_repodata):
    dpath = (module, api, cname, platform)
    need_indices = hasattr(module, "get_object")
    skip_platform = config.platforms and platform not in config.platforms

    printed = False
    if src_pkgs and all(k.get("_skip") for k in src_pkgs.values()):
        _l.info("| all packages filtered")
        printed = True
    if not src_pkgs and not dst_pkgs:
        if skip_platform:
            _l.info("| skipping platform")
            printed = True
        else:
            _l.info("| no packages")
            printed = True

    # Sort by ascending package name, then descending version, then filename.
    # This helps to ensure that the newest version of a package furnishes
    # its data to the repository. To accomplish this we rely on the stability
    # of the sorted() function and work backwards.
    # First, sort by package name
    sorted_keys = sorted(set(src_pkgs) | set(dst_pkgs))
    # Second, sort by reverse version order
    sorted_keys = sorted(
        sorted_keys,
        key=lambda x: VersionOrder(
            (src_pkgs.get(x) or dst_pkgs.get(x)).get("version", "0")
        ),
        reverse=True,
    )
    # Finally, sort by package name
    sorted_keys = sorted(
        sorted_keys,
        key=lambda x: (src_pkgs.get(x) or dst_pkgs.get(x)).get("name", x),
    )

    pkg_sets = sync_files(
        module,
        api,
        cname,
        platform,
        sorted_keys,
        src_pkgs,
        dst_pkgs,
        ignore_missing=config.ignore_missing,
        can_patch=need_indices or hasattr(module, "patch_package"),
        printed=printed,
    )

    if dst_repodata.get("extras"):
        dst_pkgs = dst_repodata.pop("extras")
        sorted_keys = sorted(dst_pkgs)
        sync_files(*dpath, sorted_keys, {}, dst_pkgs, printed=True)

    if need_indices:
        dst_repodata["packages"] = {
            k: v for k, v in dst_pkgs.items() if k.endswith(".tar.bz2")
        }
        dst_repodata["packages.conda"] = {
            k: v for k, v in dst_pkgs.items() if k.endswith(".conda")
        }
        pkg_sets["_repodata"] = dst_repodata
        dst_pkgs = dst_repodata.pop("indices", {})
        src_pkgs = build_subdir_indices(cname, platform, dst_repodata)
        sorted_keys = sorted(set(src_pkgs) | set(dst_pkgs))
        sync_files(*dpath, sorted_keys, src_pkgs, dst_pkgs, printed=True)

    return pkg_sets


def sync_channel(module, api, cname):
    summary_data = {}
    try:
        _call("begin_channel", module, api, cname, dry_run=config.dry_run)
    except MirrorException:
        config.error_count += 1
        return
    try:
        all_sources = retrieve_sources(module, api, cname)
    except MirrorException:
        config.error_count += 1
        return
    try:
        filter_all(config, all_sources)
    except MirrorException:
        config.error_count += 1
        return
    try:
        all_destinations = retrieve_destinations(module, api, cname)
    except MirrorException:
        config.error_count += 1
        return
    _l.info("Synchronizing")
    need_indices = False
    with log_prefix("  "):
        repodatas = {}
        sync_platforms = known_subdirs() if config.clean else all_sources
        for platform in sync_platforms:
            _l.info("Platform: %s", platform)
            with log_prefix("  "):
                src_pkgs = all_sources.get(platform) or {}
                dst_pkgs, dst_repodata = all_destinations[platform]
                try:
                    result = sync_platform(
                        module, api, cname, platform, src_pkgs, dst_pkgs, dst_repodata
                    )
                except MirrorException:
                    config.error_count += 1
                    result = {}
            if "_repodata" in result:
                repodatas[platform] = result.pop("_repodata")
                need_indices = True
            if any(result.values()):
                summary_data[platform] = result
            if config.error_count >= config.max_failures:
                break
        if need_indices:
            _l.info("Indices:")
            with log_prefix("  "):
                try:
                    dst_pkgs = _call("list_objects", module, api, cname) or {}
                    src_pkgs = build_channel_indices(cname, repodatas)
                    skeys = sorted(set(src_pkgs) | set(dst_pkgs))
                    sync_files(module, api, cname, None, skeys, src_pkgs, dst_pkgs)
                except MirrorException:
                    config.error_count += 1
    _call("finish_channel", module, api, cname, dry_run=config.dry_run)
    return summary_data


def get_api(module):
    api = _call("get_repo_api", module, config)
    valid_formats = getattr(config, "valid_formats", ("conda", "tarbz2"))
    if "conda" not in valid_formats:
        if config.format_policy in ("only-conda", "transmute-conda"):
            return "Format policy '%s' not supported" % config.format_policy
        elif config.format_policy != "only-tarbz2":
            config.format_policy = "transmute-tarbz2"
    elif "tarbz2" not in valid_formats:
        if config.format_policy in ("only-tarbz2", "transmute-bz2"):
            return "Format policy '%s' not supported" % config.format_policy
        elif config.format_policy != "only-conda":
            config.format_policy = "transmute-conda"
    elif config.format_policy == "default":
        config.format_policy = "prefer-conda"
    return api


def print_config(config):
    _l.info("")
    _l.info("Recipe")
    _l.info("-------")
    if getattr(config, "path", None):
        _l.info("Config file: %s", config.path)
    cname = cfull = config.dest_channel
    if config.dest_channel_alias:
        cfull = config.dest_channel_alias.rstrip("/") + "/" + cname
        if cfull.startswith("file://"):
            cfull = cfull[7:]
    elif config.dest_site:
        cfull = config.dest_site.rstrip("/") + "/" + cname
    _l.info("Destination: %s", cfull)
    if getattr(config, "access_policy", None):
        _l.info("Access policy: %s", config.access_policy)
    _l.info("Format policy: %s", config.format_policy)
    _l.info("Source%s:" % ("" if len(config.channels) == 1 else "s"))
    for c in config.channels:
        _l.info("  - %s", c)
    if config.platforms:
        _l.info("Subdirs:")
        for p in config.platforms:
            _l.info("  - %s", p)
    else:
        _l.info("Subdirs: <all>")
    if (
        config.python_versions
        or config.r_versions
        or config.pkg_list
        or config.license_exclusions
        or config.constraints
        or config.exclusions
        or config.inclusions
    ):
        _l.info("Filtering:")
        if config.python_versions:
            _l.info("  Python versions: %s", "|".join(config.python_versions))
        if config.r_versions:
            _l.info("  R versions: %s", "|".join(config.r_versions))
        if config.pkg_list:
            _l.info("  Include only:")
            for p in config.pkg_list:
                _l.info("  - %s" % p)
            _l.info("  Dependencies?: %s", "yes" if config.dependencies else "no")
        _l.info("  Prune uninstallable?: %s", "yes" if config.pruning else "no")
        if config.license_exclusions:
            _l.info("  License exclusions:")
            for p in config.license_exclusions:
                _l.info("  - %s" % p)
        if config.constraints:
            _l.info("  Constraints:")
            for p in config.constraints:
                _l.info("  - %s" % p)
        if config.exclusions:
            _l.info("  Exclusions:")
            for p in config.exclusions:
                _l.info("  - %s" % p)
        if config.inclusions:
            _l.info("  Inclusions:")
            for p in config.inclusions:
                _l.info("  - %s" % p)
    else:
        _l.info("Filters: <none>")
    if config.ignore_missing:
        _l.info("Ignore missing: yes")
    ssls = config.verify_ssl_mode("src")
    ssld = config.verify_ssl_mode("dst")
    if not (ssls is True and ssld is True):
        _l.info("SSL verification:")
        _l.info(
            "  Source: %s",
            "disabled"
            if not ssls
            else "CA file"
            if isinstance(ssls, str)
            else "enabled",
        )
        _l.info(
            "  Destination: %s",
            "disabled"
            if not ssld
            else "custom CA file"
            if isinstance(ssld, str)
            else "enabled",
        )
    _l.info("Clean: %s", "yes" if config.clean else "no")
    _l.info("Verbose: %s", "yes" if config.verbose else "no")
    _l.info("Dry run: %s", "yes (no changes will be made)" if config.dry_run else "no")
    _l.info("Connection attempts: %d", config.max_attempts)
    _l.info("Tolerated failures: %d", config.max_failures)


VALUE_STR = {
    "before": "before mirror",
    "kept": "preserved due to clean=false",
    "replaced": "replaced for content changes",
    "patched": "patched for metadata changes",
    "filtered": "blocked due to filtering",
    "after": "after mirror",
    "skipped": "skipped due to format policy",
    "blocked": "blocked due to format policy",
    "errored": "FAILURES encountered",
}


ALL_VALS = (
    "before",
    "added",
    "removed",
    "kept",
    "replaced",
    "patched",
    "filtered",
    "skipped",
    "blocked",
    "errored",
    "after",
)


def print_summary(config, summary_data):
    if summary_data is None:
        _l.warning("| could not be completed")
        return
    elif not summary_data:
        _l.info("| no actions taken")
        return
    strlen = max(
        len(str(v)) for pkg_counts in summary_data.values() for v in pkg_counts.values()
    )
    for platform, pkg_counts in summary_data.items():
        _l.info("Platform: %s", platform)
        with log_prefix("  "):
            for key in ALL_VALS:
                if key in ("before", "after") or pkg_counts.get(key, 0):
                    _l.info("%*d %s", strlen, pkg_counts[key], VALUE_STR.get(key, key))


def main(module):
    parse_argv()
    set_log_level(config.verbose)
    start_time = datetime.datetime.now()
    _l.info("START: %s", start_time.isoformat())
    config.error_count = 0
    try:
        api = get_api(module)
        print_config(config)
        _l.info("")
        _l.info("Actions")
        _l.info("-------")
        cname = config.dest_channel
        summary_data = {}
        try:
            summary_data[cname] = sync_channel(module, api, cname)
        except MirrorException as exc:
            _l.error(exc)
            config.error_count += 1
        _l.info("")
        _l.info("Summary")
        _l.info("-------")
        for cname, sdict in summary_data.items():
            _l.info("Channel: %s", cname)
            with log_prefix("  "):
                print_summary(config, sdict)
    except MirrorException:
        config.error_count += 1
    except Exception:
        _l.info("---------")
        _l.error(traceback.format_exc())
        _l.info("---------")
        config.error_count = -1
    _l.info("")
    if config.error_count:
        _l.info("Sync operation DID NOT complete successfully.")
        _l.info("Please examine the logs for more details.")
    elif config.dry_run:
        _l.info("REMINDER: this was a dry run. The 'final' numbers")
        _l.info("above represent what *would* have occurred after")
        _l.info("a successful, actual run.")
    else:
        _l.info("Sync operation completed successfully.")
    end_time = datetime.datetime.now()
    _l.info("")
    _l.info("END: %s (%s)", end_time.isoformat(), end_time - start_time)
    return config.error_count



import hashlib
import contextlib
import re
import socket


@contextlib.contextmanager
def fp_or_bytes(path_or_data):
    if isinstance(path_or_data, bytes):
        yield path_or_data
    else:
        with open(path_or_data, "rb") as fp:
            yield fp


def compute_checksums(path_or_data):
    h_md5 = hashlib.md5()
    h_sha1 = hashlib.sha1()
    h_sha256 = hashlib.sha256()
    if isinstance(path_or_data, bytes):
        h_md5.update(path_or_data)
        h_sha1.update(path_or_data)
        h_sha256.update(path_or_data)
        sz = len(path_or_data)
    else:
        sz = 0
        chunk = None
        with open(path_or_data, "rb") as fp:
            while chunk != b"":
                chunk = fp.read(1024)
                sz += len(chunk)
                h_md5.update(chunk)
                h_sha1.update(chunk)
                h_sha256.update(chunk)
    result = {
        "size": sz,
        "md5": h_md5.hexdigest(),
        "sha256": h_sha256.hexdigest(),
        "sha1": h_sha1.hexdigest(),
    }
    return result


def _drop_sfx(x):
    if x.endswith(".conda"):
        return x[:-6]
    if x.endswith(".tar.bz2"):
        return x[:-8]
    return x


def count_pkgs(pdict, filter=None):
    n_pkgs = sum(filter is None or filter(k, v) for k, v in pdict.items())
    n_blobs = len(
        set(_drop_sfx(k) for k, v in pdict.items() if filter is None or filter(k, v))
    )
    return n_pkgs, n_blobs


def site_to_url(ds, api_suffix=None):
    """
    Attempts to intelligently determine if a dest_site value can be mapped
    to a full URL. While not extremely strict, but it does these things:
    - If a valid http or https URL is supplied, it returns it unchanged
    - It confirms that the netloc is formatted as a a valid IPv4 address,
      IPv6 address, domain name, or "localhost"
    - It confirms that the port, if supplied, is an integer from 1-65536
    If these conditions are satisfied, it returns a URL, adding http://
    or https:// as appropriate. Otherwise it returns None. Examples:
    - 127.0.0.1 -> http://127.0.0.1
    - 127.0.0.1:443 -> https://127.0.0.1:443
    - anaconda.com/api -> https://anaconda.com/api
    - https://anaconda.com:84x3/api -> None
    """

    def _cleanup(url, secure):
        # Add scheme if missing
        # Strip any trailing slash
        # If the api suffix is missing, append it
        # If the api suffix is present, strip additional path elements
        if not url.startswith(("http://", "https://")):
            pfx = "https://" if secure else "http://"
            url = pfx + url
        url = url.rstrip("/")
        if api_suffix:
            sfx_s = "/" + api_suffix.lstrip("/")
            if not url.endswith(sfx_s):
                sfx_t = sfx_s + "/"
                if sfx_t in url:
                    url = url[: url.find(sfx_t) + len(sfx_s)]
                else:
                    url += sfx_s
        return url

    # Extract and normalize the netloc (which might include a port)
    _, netloc, _ = re.match("^(https?://)?([^/]+)(.*)$", ds).groups()
    netloc = netloc.encode("idna").decode("ascii").lower()

    # First check to see if it is a valid ipv6 host
    # without a trailing port value
    try:
        socket.inet_pton(socket.AF_INET6, netloc)
        return _cleanup(ds, False)
    except Exception:
        pass

    # If a port value is present, strip it off and check validity
    if ":" in netloc:
        netloc, port = netloc.rsplit(":", 1)
        try:
            port = int(port)
        except ValueError:
            return
        if port is None or not (0 < port < 65536):
            return
    else:
        port = None

    # A single trailing period is allowed
    if netloc.endswith("."):
        netloc = netloc[:-1]
    if not (0 < len(netloc) <= 255):
        return

    # Treat local hosts specially; default to http
    if netloc in ("localhost", "127.0.0.1", "0.0.0.0"):
        return _cleanup(ds, port in (443, 8443))

    # Check for valid domain names
    if not re.match(r"^([0-9a-z][-\w]*[0-9a-z]\.)+[a-z0-9\-]{2,15}$", netloc):
        # Check for IPv6 again with the port stripped off
        try:
            socket.inet_pton(socket.AF_INET6, netloc)
        except Exception:
            # Check for IPv4
            try:
                socket.inet_pton(socket.AF_INET, netloc)
            except Exception:
                return

    # Default to https unless port is 80, 8080, or 8088
    return _cleanup(ds, port not in (80, 8080, 8088))


class MirrorException(Exception):
    def __init__(self, msg):
        self.message = msg.lstrip()



import re
import string

from conda.models.match_spec import MatchSpec

from .logging import logger as _l, log_prefix
from .utils import MirrorException


standard_license_families = {
    "AGPL",
    "APACHE",
    "BSD",
    "CC",
    "GPL",
    "GPL2",
    "GPL3",
    "GPL-GCC",
    "LGPL",
    "MIT",
    "MOZILLA",
    "NONE",
    "OTHER",
    "PROPRIETARY",
    "PSF",
    "PUBLICDOMAIN",
}

_matchspec_cache = {}


def _matchspec(dstr):
    """
    Return a MatchSpec, a normalized MatchSpec string,
    and the package name for a raw MatchSpec input string.
    Use caching to improve performance
    """
    drec = _matchspec_cache.get(dstr)
    if drec is None:
        dep = MatchSpec(dstr)
        drec = (dep, str(dep), dep.name)
        _matchspec_cache[dstr] = _matchspec_cache[drec[1]] = drec
    return drec


def _matchname(name):
    return _matchspec(name)[0]._match_components["name"]


GLOB = _matchname("*")


def filter_format(config, upstream_indices):
    counts = {}
    if config.format_policy == "keep-both" or not upstream_indices:
        return counts
    mode, fmt = config.format_policy.split("-", 1)
    if fmt == "tarbz2":
        good_fmt, bad_fmt = ".tar.bz2", ".conda"
    else:
        good_fmt, bad_fmt = ".conda", ".tar.bz2"
    n_bad = len(bad_fmt)
    mode, fmt = config.format_policy.split("-", 1)
    for platform, upstream_index in upstream_indices.items():
        new_pkgs = {}
        t_counts = {"redundant": 0, "dropped": 0, "transforming": 0}
        for fn, info in upstream_index.items():
            if fn.endswith(good_fmt):
                continue
            good_fn = fn[:-n_bad] + good_fmt
            if good_fn in upstream_index:
                t_counts["redundant"] += 1
                info["_skip"] = "redundant"
            elif info.get("_skip"):
                pass
            elif mode == "only":
                t_counts["redundant"] += 1
                info["_skip"] = "format"
            elif mode == "transmute":
                # For transmuting, we actually need to add the
                # destination filename to the package list. We
                # copy the source record to that new filename.
                t_counts["transforming"] += 1
                info_dst = info.copy()
                info_dst["fn"] = fn
                info["_skip"] = "transmuting"
                new_pkgs[good_fn] = info_dst
        upstream_index.update(new_pkgs)
        if any(t_counts.values()):
            counts[platform] = ", ".join(
                "%d %s" % (v, k) for k, v in t_counts.items() if v
            )
    return counts


class MatchBase(object):
    @classmethod
    def from_matchspec(cls, ms):
        if isinstance(ms, str):
            ms, _, _ = _matchspec(ms)
        name = ms._match_components["name"]
        children = [
            MatchField(k, v) for k, v in ms._match_components.items() if k != "name"
        ]
        return name, MatchAnd.from_specs(*children)

    def __init__(self):
        self._hash = None

    def __hash__(self):
        if self._hash is None:
            items = [(k, v) for k, v in self.__dict__.items() if not k.startswith("_")]
            self._hash = hash(tuple(sorted(items)))
        return self._hash

    def __eq__(self, other):
        return type(self) == type(other) and (  # noqa
            self is other or hash(self) == hash(other)
        )

    def match(self, rec):
        raise NotImplementedError()

    def match_callback(self, rec, if_true=None, if_false=None):
        result = self.match(rec)
        callback = if_true if result else if_false
        if callback is not None:
            callback(self, rec)
        return result


class MatchNone(MatchBase):
    def match(self, rec):
        return False

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "<none>"

    def __repr__(self):
        return "MatchNone()"


class MatchAny(MatchBase):
    def match(self, rec):
        return True

    def __str__(self):
        return "<any>"

    def __repr__(self):
        return "MatchAny()"


def _normalize_license_name(s, license):
    if not s:
        s = license
    if not s:
        return "NONE"
    if "GPL" in s and "GCC-exception" in license:
        return "GPL-GCC"
    if "LGPL" in s or "LGPL" in license:
        return "LGPL"
    s = s.upper().strip()
    s = re.sub(r"GENERAL PUBLIC LICENSE", "GPL", s)
    s = re.sub(r"LESSER\s+", "L", s)
    s = re.sub(r"AFFERO\s+", "A", s)
    s = re.sub("[%s]" % re.escape(string.punctuation), "", s)
    s = re.sub(r"\s+", "", s)
    return s.strip() or "NONE"


class MatchLicense(MatchBase):
    def __init__(self, license):
        super().__init__()
        self.license = license

    def __str__(self):
        return f"license({self.license})"

    def __repr__(self):
        return f"MatchLicense({repr(self.license)})"

    def match(self, rec):
        license_fam = _normalize_license_name(
            rec.get("license_family"), rec.get("license")
        )
        if license_fam not in standard_license_families:
            for fam in standard_license_families:
                if fam in license_fam:
                    license_fam = fam
                    break
            else:
                if re.match(r"(^|[^AL])GPL([^23]*2|.*>= *)2", license_fam):
                    license_fam = "GPL2"
                elif re.match(r"(^|[^AL])GPL([^23]*3|.*>= *)3", license_fam):
                    license_fam = "GPL3"
        return self.license == license_fam


class MatchDepVer(MatchBase):
    def __init__(self, dname, dversion):
        super().__init__()
        self.dname = dname
        self.dversion = dversion
        self._prototypes = [
            {
                "name": dname,
                "version": dversion + "." + str(mv),
                "build": "test",
                "build_number": 0,
            }
            for mv in (0, 1)
        ]

    def __repr__(self):
        return f"MatchDepVer({repr(self.dname)},{repr(self.dversion)})"

    def __str__(self):
        return f"depends({self.dname}={self.dversion})"

    def match(self, rec):
        if rec["name"] == self.dname:
            return True
        for field in ("depends", "constrains"):
            for d in rec.get(field) or ():
                if d.startswith(self.dname):
                    if d == self.dname:
                        return True
                    ms, _, msname = _matchspec(d)
                    if msname != self.dname:
                        continue
                    return any(ms.match(prec) for prec in self._prototypes)
        return True


class MatchField(MatchBase):
    def __init__(self, field, matcher):
        super().__init__()
        self.field = field
        self.matcher = matcher

    def __str__(self):
        return f"{self.field}:{str(self.matcher)}"

    def __repr__(self):
        return f"MatchField({repr(self.field)},{repr(self.matcher)})"

    def match(self, rec):
        return self.field in rec and self.matcher.match(rec[self.field])


class MatchNot(MatchBase):
    @classmethod
    def from_specs(cls, arg):
        if isinstance(arg, (str, MatchSpec)):
            _, arg = MatchBase.from_matchspec(arg)
        if isinstance(arg, MatchAny):
            return MatchNone()
        if isinstance(arg, MatchNone):
            return MatchAny()
        if isinstance(arg, MatchNot):
            return arg.atom
        return MatchNot(arg)

    def __init__(self, atom):
        super().__init__()
        self.atom = atom

    def __repr__(self):
        return f"MatchNot({repr(self.atom)})"

    def __str__(self):
        return f"not({str(self.atom)})"

    def match(self, rec):
        return not self.atom.match(rec)

    def match_callback(self, rec, if_true=None, if_false=None):
        return not self.atom.match_callback(rec, if_true=if_false, if_false=if_true)


class MatchTree(MatchBase):
    oper = "<none>"

    @classmethod
    def from_specs(cls, *args, internal=False):
        no_op = MatchAny if cls is MatchAnd else MatchNone
        new_args = set()
        for arg in args:
            if arg is None or isinstance(arg, no_op):
                continue
            elif isinstance(arg, (str, MatchSpec)):
                _, arg = MatchBase.from_matchspec(arg)
            elif not isinstance(arg, MatchBase):
                raise TypeError(f"atom must not be a(n) {repr(arg)}")
            if isinstance(arg, MatchTree) and cls.oper == arg.oper:
                # associativity
                new_args.update(arg.atoms)
            elif (
                isinstance(arg, MatchNot)
                and isinstance(arg.atom, MatchTree)
                and cls.oper != arg.atom.oper
            ):
                # demorgans
                new_args.update(MatchNot(a) for a in arg.atom.atoms)
            elif isinstance(arg, MatchAny):
                if cls.oper == "or":
                    new_args = {arg}
                    break
            elif isinstance(arg, MatchNone):
                if cls.oper == "and":
                    new_args = {arg}
                    break
            else:
                new_args.add(arg)
        if internal:
            return new_args
        elif len(new_args) > 1:
            return cls(*new_args, internal=True)
        elif len(new_args) == 1:
            return next(iter(new_args))
        elif cls == MatchAnd:
            return MatchAny()
        else:
            return MatchNone()

    def __init__(self, *args, internal=False):
        super().__init__()
        if not internal:
            args = self.from_specs(*args, internal=True)
        self.atoms = args

    def __repr__(self):
        return f"{self.__class__.__name__}({','.join(map(repr,self.atoms))})"

    def __str__(self):
        return f"{self.oper}({','.join(map(str,self.atoms))})"


class MatchAnd(MatchTree):
    oper = "and"

    def match(self, rec):
        return all(a.match(rec) for a in self.atoms)

    def match_callback(self, rec, if_true=None, if_false=None):
        matched = True
        for a in self.atoms:
            if not a.match_callback(rec, if_false=if_false):
                matched = False
        if matched and if_true is not None:
            if_true(self, rec)
        return matched


class MatchOr(MatchTree):
    oper = "or"

    def match(self, rec):
        return any(a.match(rec) for a in self.atoms)

    def match_callback(self, rec, if_true=None, if_false=None):
        matched = False
        for a in self.atoms:
            if a.match_callback(rec, if_true=if_true):
                matched = True
        if not matched and if_false is not None:
            if_false(self, rec)
        return matched


def filter_spec_dict(matches):
    filters = {}
    for ms in matches:
        msname, matcher = MatchBase.from_matchspec(ms)
        filters[msname] = MatchOr.from_specs(filters.get(msname), matcher)
    return filters


def filter_version_dict(versions, package):
    constraints = {}
    vlist = [v.rstrip("*").rstrip(".") for v in versions]
    constraints[_matchname(package)] = MatchOr.from_specs(
        *(f"{package}={v}.*" for v in vlist)
    )
    dspec = MatchOr.from_specs(*(MatchDepVer(package, v) for v in vlist))
    constraints[GLOB] = MatchAnd.from_specs(constraints.get(GLOB), dspec)
    return constraints


def filter_license_dict(licenses):
    exclusions = {}
    lspecs = [MatchLicense(lic) for lic in licenses]
    lspec = MatchOr.from_specs(*lspecs)
    exclusions[GLOB] = MatchAnd.from_specs(exclusions.get(GLOB), lspec)
    return exclusions


def filter_requests(groups, requests, prune=False):
    touched_names = set()
    for msname, matcher in requests.items():
        for name in filter(msname.match, groups):
            touched_names.add(name)
            for version, sdict in groups[name].items():
                for subdir, pdict in sdict.items():
                    for fn, pkg in pdict.items():
                        if matcher.match(pkg):
                            pkg["_touched"] = True
    prune_limit = None if prune else touched_names
    prune_untouched(groups, "pkg_list", prune_limit)
    return touched_names


def filter_spec(groups, filter_dict, positive, reason):
    if not isinstance(filter_dict, dict):
        filters, filter_dict = filter_dict, {}
        for ms in filters:
            mname, ms = MatchBase.from_matchspec(ms)
            filter_dict[mname] = MatchOr.from_specs(filter_dict.get(mname), ms)
    counts = {}
    msname = ""

    def _count(matcher, pkg):
        r = f"{msname}:{matcher}"
        name = pkg["name"]
        crec = counts.setdefault(r, {})
        crec[name] = crec.get(name, 0) + 1

    for msname, matcher in filter_dict.items():
        for name in filter(msname.match, groups):
            for version, sdict in groups[name].items():
                for subdir, pdict in sdict.items():
                    for fn, pkg in pdict.items():
                        _skip = pkg.get("_skip")
                        if _skip and not _skip.startswith("filter:"):
                            continue
                        if not reason and not _skip:
                            continue
                        kwargs = {"if_true" if positive else "if_false": _count}
                        if bool(matcher.match_callback(pkg, **kwargs)) == positive:
                            if reason is None:
                                del pkg["_skip"]
                            else:
                                pkg["_skip"] = f"filter:{reason}"
    return counts


def prune_untouched(groups, reason=None, names=None):
    for name in names or list(groups):
        vdict = groups[name]
        for version, sdict in list(vdict.items()):
            for subdir, pdict in list(sdict.items()):
                for fn, pkg in list(pdict.items()):
                    if pkg.get("_touched"):
                        del pkg["_touched"]
                    elif reason is not None:
                        pkg.setdefault("_skip", reason)
                    if pkg.get("_skip"):
                        del pdict[fn]
                if not pdict:
                    del sdict[subdir]
            if not sdict:
                del vdict[version]
        if not vdict:
            del groups[name]


def prune_uninstallable(groups):
    """Remove packages from the database that are missing at least one
    valid dependency. This should in theory produce a set of packages
    that can be installed singly. But this is a conservative pruning;
    dependency cycles and noarch packages are treated in such a way
    that some uninstallable packages may remain."""

    match_cache = {}

    def _test_dep(top_sub, dstr):
        # Assume virtual packages are available
        if dstr.startswith("__"):
            return True
        dep, dstr, dname = _matchspec(dstr)
        dkey = (top_sub, dstr)
        if dkey in match_cache:
            return match_cache[dkey]
        match_cache[dkey] = True
        dversion = None if dep is None else dep.version
        for version, sdict in groups.get(dname, {}).items():
            if not (dversion is None or dversion.match(version)):
                continue
            for subdir in (top_sub, "noarch"):
                for fn, pkg in sdict.get(subdir, {}).items():
                    if dep is None or dep.match(pkg):
                        if _touch(top_sub, subdir, fn, pkg):
                            match_cache[dkey] = True
                            return True
        match_cache[dkey] = False
        return False

    pkg_touched = {}

    def _touch(top_sub, sub, fn, pkg):
        pkey = (top_sub, sub, fn)
        if pkey in pkg_touched:
            return pkg_touched[pkey]
        # This breaks potential cycles in the dep graph.
        # Packages that are uninstallable despite this
        # assumption will still be marked False below.
        pkg_touched[pkey] = True
        result = all(_test_dep(top_sub, dstr) for dstr in pkg["depends"])
        pkg_touched[pkey] = result
        if result:
            pkg["_touched"] = True
        return result

    # All of the platforms we need to search over, minus noarch
    platforms = sorted(
        set(
            subdir
            for vdict in groups.values()
            for sdict in vdict.values()
            for subdir in sdict
        )
        - {"noarch"}
    )

    # Iterate through each the package and check it for individual
    # installability. Some packages may prove installable on some
    # platforms, but not others. prune_untouched groups the platform
    # tests together so such packages will be preserved.
    for name, vdict in groups.items():
        for version, sdict in vdict.items():
            for subdir, pdict in sdict.items():
                top_subs = platforms if subdir == "noarch" else (subdir,)
                for fn, pkg in pdict.items():
                    for top_sub in top_subs:
                        _touch(top_sub, subdir, fn, pkg)

    prune_untouched(groups, "uninstallable")


def prune_untouchables(groups, requests, reason="unreachable"):
    """Find all packages that can be reached from the set of
    requested dependencies, and then prune all others."""

    # Keeps track of the packages we have already studied. This saves
    # not just the package itself, but the target subdir it is being
    # installed in ("top_sub"). This is because the dependency graph
    # for noarch packages will necessarily be platform dependent
    pkg_touched = set()

    # Keeps track of the dependency searches we have already performed
    match_touched = set()

    def _touch(subdir, fn, pkg, top_sub, dep=None):
        key = (top_sub, subdir, fn)
        if key in pkg_touched or dep is not None and not dep.match(pkg):
            return
        pkg_touched.add(key)
        pkg["_touched"] = True
        l_deps = pkg["depends"] + ["pip"] if pkg["name"] == "python" else pkg["depends"]
        for dstr in l_deps:
            dep, dstr, dname = _matchspec(dstr)
            dkey = (top_sub, dstr)
            if dkey not in match_touched:
                match_touched.add(dkey)
                dversion = dep.version
                for version, sdict in groups.get(dname, {}).items():
                    if dversion is None or dversion.match(version):
                        for subdir2 in (top_sub, "noarch"):
                            for fn2, pkg2 in sdict.get(subdir2, {}).items():
                                _touch(subdir2, fn2, pkg2, top_sub, dep)

    all_subdirs = sorted(
        set(
            subdir
            for vdict in groups.values()
            for sdict in vdict.values()
            for subdir in sdict
        )
        - {"noarch"}
    )
    if requests is None:
        requests = groups
    for name in requests:
        for version, sdict in groups.get(name, {}).items():
            for subdir, pdict in sdict.items():
                try_sub = all_subdirs if subdir == "noarch" else (subdir,)
                for fn, pkg in pdict.items():
                    for subdir2 in try_sub:
                        _touch(subdir, fn, pkg, subdir2)

    prune_untouched(groups, reason)


def count_groups(groups, limit=None):
    n_names = n_versions = n_packages = 0
    for name, vdict in groups.items():
        if limit is not None and name not in limit:
            continue
        n_names += 1
        n_versions += len(vdict)
        for sdict in vdict.values():
            for pdict in sdict.values():
                n_packages += len(pdict)
    return n_names, n_versions, n_packages


def repodata_to_groups(upstream_indices):
    groups = {}
    for subdir, data in upstream_indices.items():
        for fn, pkg in data.items():
            if pkg.get("_skip"):
                continue
            groups.setdefault(pkg["name"], {}).setdefault(
                pkg["version"], {}
            ).setdefault(subdir, {})[fn] = pkg
    return groups


def filter_all(config, upstream_indices):
    def _info(groups, why):
        why = f"{why}: " if why.strip() else why
        n_names, n_versions, n_packages2 = count_groups(groups)
        _l.info(f"{why}{n_names} names / {n_versions} versions / {n_packages2} blobs")
        return n_packages2

    _l.info("Applying filters")
    with log_prefix("  | "):
        groups = repodata_to_groups(upstream_indices)
        n_pkgs = _info(groups, "initial")

        if config.pkg_list:
            requests = filter_spec_dict(config.pkg_list)
            request_names = filter_requests(
                groups, requests, prune=not config.dependencies
            )
            t_groups = {k: v for k, v in groups.items() if k in request_names}
            n_pkgs = _info(t_groups, "pkg_list")
            if config.dependencies:
                prune_untouchables(groups, request_names, "pkg_list")
                t_groups = {k: v for k, v in groups.items() if k not in request_names}
                _info(t_groups, "dependencies")
                n_pkgs = _info(groups, "total")

        if config.constraints:
            constraints = filter_spec_dict(config.constraints)
            counts = filter_spec(groups, constraints, False, "constraints")
            _info(groups, "constraints")
            for reason, count in counts.items():
                _l.info(f"  {reason}: {count}")

        if config.python_versions:
            constraints = filter_version_dict(config.python_versions, "python")
            counts = filter_spec(groups, constraints, False, "python_version")
            _info(groups, "python versions")
            for reason, count in counts.items():
                _l.info(f"  {reason}: {sum(count.values())}")

        if config.r_versions:
            constraints = filter_version_dict(config.python_versions, "r-base")
            counts = filter_spec(groups, constraints, False, "r_version")
            _info(groups, "R versions")
            for reason, count in counts.items():
                _l.info(f"  {reason}: {sum(count.values())}")

        if config.license_exclusions:
            exclusions = filter_license_dict(config.license_exclusions)
            counts = filter_spec(groups, exclusions, True, "license")
            _info(groups, "license")
            for reason, count in counts.items():
                _l.info(f"  {reason}: {', '.join(count)}")

        if config.exclusions:
            exclusions = filter_spec_dict(config.exclusions)
            counts = filter_spec(groups, exclusions, True, "exclusions")
            _info(groups, "exclusions")
            for reason, count in counts.items():
                _l.info(f"  {reason}: {sum(count.values())}")

        if config.inclusions:
            inclusions = filter_spec_dict(config.inclusions)
            counts = filter_spec(groups, inclusions, True, None)
            _info(groups, "inclusions")
            for reason, count in counts.items():
                _l.info(f"  {reason}: {sum(count.values())}")

        counts = filter_format(config, upstream_indices)
        # This rolls up all of the packages eliminated from
        # the filters above as well
        groups = repodata_to_groups(upstream_indices)
        if counts:
            _info(groups, "format policy")
            for platform, count in counts.items():
                _l.info(f"  {platform}: {count}")

        if config.pruning:
            if not config.pkg_list:
                request_names = groups.keys()
            n_iter = 1
            while True:
                prune_uninstallable(groups)
                n_pkgs2 = _info(groups, "pruning uninstallable %d" % n_iter)
                prune_untouchables(groups, request_names)
                if n_iter > 1 and n_pkgs2 == n_pkgs:
                    break
                n_pkgs3 = _info(groups, "pruning unreachable %d" % n_iter)
                if n_pkgs3 == n_pkgs or n_iter == 1 and n_pkgs3 == n_pkgs2:
                    break
                n_pkgs = n_pkgs3
                n_iter += 1

    if config.pkg_list:
        missing = set()
        groups = repodata_to_groups(upstream_indices)
        for msname, msmatch in requests.items():
            for name in groups:
                if msname.match(name) and any(
                    msmatch.match(pkg)
                    for vdict in groups[name].values()
                    for sdict in vdict.values()
                    for pkg in sdict.values()
                ):
                    break
            else:
                missing.add(f"{msname}: {msmatch}")
        if missing:
            msg = [
                "One or more requested packages were found to be uninstallable:",
                *(" - " + r for r in sorted(missing)),
                "Please examine your filter and try again.",
            ]
            for line in msg:
                _l.error("! %s", line)
            raise MirrorException("\n".join(msg))



import sys

from .utils import MirrorException


def get_repo_api(config):
    global parse_conda
    try:
        from anaconda_enterprise.cli.parse import parse_conda
        from anaconda_enterprise.cli import config as ap_config
        from anaconda_enterprise.cli import api as ap_api
    except ImportError:
        raise MirrorException("Please install the 'anaconda-enterprise-cli' package.")
    ap_config.config.load_schema()
    ap_config.config.load_files()
    ap_config.config.component = [
        "sites",
        config.dest_site or ap_config.config["default_site"],
    ]
    ap_tokens = ap_api.TokenStore(ap_config.config.token_path)
    api = ap_api.ApiClient(ap_config.config, ap_tokens)
    if config.verify_ssl_dst:
        config.verify_ssl_dst = ap_config.config["ssl_verify"]
    ssl_verify = config.verify_ssl_mode("dst")
    api.session.verify = api.file_session.verify = ssl_verify
    # The anaconda-enterprise-cli api object provides a retry capability
    # so we can simply feed it our configurable max_attempts value
    ap_api.RETRIES = config.max_attempts or 5
    api.check()
    config.dest_channel_alias = api.make_absolute(api.links["conda"].url)
    config.valid_formats = ("tarbz2",)
    return api


def begin_channel(api, cname, dry_run):
    for channel in api._json_api("channels"):
        if channel["attributes"]["name"] == cname:
            api.channel_record = channel
            return
    raise MirrorException("Channel not found: %s" % cname)


def del_object(api, cname, platform, fn, metadata=None):
    try:
        api._json_api("files/" + metadata["extra"]["id"], method="DELETE")
    except Exception as exc:
        if "404 (Not Found)" not in str(exc):
            raise MirrorException(str(exc))


def put_object(api, cname, platform, fn, path_or_data=None, metadata=None):
    attrs = parse_conda(path_or_data)
    adict = metadata.copy()
    adict["fn"] = fn
    attrs["index-attributes"] = adict
    try:
        response = api.stage_file(api.channel_record, attrs)
        if "has-icon" in attrs["recipe-attributes"]:
            api.upload_icon(response, path_or_data)
        with open(path_or_data, "rb") as fd:
            api.upload_file(response, fd)
    except Exception as exc:
        message = getattr(exc, "message", None) or str(exc)
        status = getattr(exc, "status", 0)
        if status == 409 or "409" in message:
            return
        if status != 0:
            message = "status %d, %s" % (status, message)
        raise MirrorException(message)


def main():
    from .common import main as _main

    _main(sys.modules[__name__])



import logging
import contextlib


logger = logging.getLogger("mirror")
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)
handler.setFormatter(logging.Formatter(fmt="%(message)s"))
logger.addHandler(handler)


__level = logging.INFO
__prefix = []


def set_log_level(level):
    global __level
    if isinstance(level, bool):
        level = logging.DEBUG if level else logging.INFO
    elif isinstance(level, str):
        n_level = getattr(logging, level.upper(), None)
        if not isinstance(n_level, int):
            raise RuntimeError("Invalid log level: %s" % level)
        level = n_level
    elif not isinstance(level, int):
        raise RuntimeError("Invalid log level: %s" % level)
    __level = level
    logger.setLevel(level)
    handler.setLevel(level)


def is_verbose():
    global __level
    return __level <= logging.DEBUG


def set_log_prefix():
    global __prefix
    global handler
    fmt = "".join(__prefix + ["%(message)s"])
    handler.setFormatter(logging.Formatter(fmt=fmt))


def push_log_prefix(pfx):
    global __prefix
    n_prefix = len(__prefix)
    __prefix.append(pfx)
    set_log_prefix()
    return n_prefix


def pop_log_prefix(count=None):
    global __prefix
    if count is None:
        count = len(__prefix) - 1
    del __prefix[count:]
    set_log_prefix()


@contextlib.contextmanager
def log_prefix(pfx):
    n_prefix = push_log_prefix(pfx)
    try:
        yield
    finally:
        pop_log_prefix(n_prefix)
